from collections import OrderedDict


class ActionSetDict(OrderedDict):
    """
    Custom dict type for instantiating item classes.
    """

    def instantiate(self, name: str, *args, **kwargs) -> None:
        self[name] = self[name](*args, **kwargs)



from typing import TypeVar

from dev_tools.meta.metaclasses import MethodCheckMetaclass
from dev_tools.template.components import BaseHTMLElement, BaseButton


class AbstractActionMetaclass(MethodCheckMetaclass):
    """
    Metaclass for verifying the existence of the render method in the class.
    """

    base_class_name = 'BaseAction'
    method_name = 'render'


class BaseAction(metaclass=AbstractActionMetaclass):
    """
    Base class for defining objects as actions. Any object may be an action if it has a render method.
    """

    def __str__(self):
        return self.render()


ActionType = TypeVar('ActionType', bound=BaseAction)


class SimpleActionElement(BaseHTMLElement, BaseAction):
    """
    Base class for simple actions presented as HTML-element.
    """

    pass


class ActionButton(BaseButton, BaseAction):
    """
    Base class for action button.
    """

    pass


class ActionLink(BaseButton, BaseAction):
    """
    Base class for action link (a).
    """

    tag = 'a'



from collections import OrderedDict
from copy import deepcopy
from typing import Any, TypeVar, List, Optional, Type

from dev_tools.meta.metaclasses import DeclarativeAttributesBaseMetaclass
from dev_tools.template.mixins import TemplateObjectMixin, HttpRequestType
from django.urls import resolve

from action_set.components import BaseAction, ActionType, ActionButton
from action_set.types import ActionSetDict


class ActionsDeclarativeMetaclass(DeclarativeAttributesBaseMetaclass):
    """
    Metaclass for collecting actions for action_set.
    """

    _attribute_base_class = BaseAction

    @classmethod
    def _contribute_to_class(mcs, class_obj: Any, declared_attributes: dict) -> None:
        class_obj.declared_actions = declared_attributes
        class_obj.base_actions = declared_attributes


class BaseActionSet(TemplateObjectMixin, metaclass=ActionsDeclarativeMetaclass):
    """
    Base raw class for creating custom Action Sets.
    """

    context_object_name = 'action_set'

    def __init__(self):
        self.actions = deepcopy(self.base_actions)

    def __getitem__(self, name):
        return self.actions[name]

    def __iter__(self):
        for _, action in self.actions.items():
            yield action


class ActionSet(BaseActionSet):
    """
    Main class for grouping actions with custom template support.
    """

    def __init__(self, request: HttpRequestType, action_set_group: Optional['ActionSetGroupType'] = None):
        super().__init__()
        self.request = request
        self.action_set_group = action_set_group

    def _get_request(self) -> Optional[HttpRequestType]:
        return self.request


class GroupedActionSet:
    """
    Mixin for adding Action Set grouper variable.
    """

    def __init__(self, action_set_group: 'ActionSetGroupType'):
        self.action_set_group = action_set_group
        request = action_set_group.request
        super().__init__(request)


ActionSetType = TypeVar('ActionSetType', bound=ActionSet)


class ActionSetDeclarativeMetaclass(DeclarativeAttributesBaseMetaclass):
    """
    Metaclass for collecting action_sets for action group.
    """

    _attribute_base_class = BaseActionSet.__class__

    @classmethod
    def _contribute_to_class(mcs, class_obj: Any, declared_attributes: dict) -> None:
        class_obj.declared_action_sets = declared_attributes
        class_obj.base_action_sets = declared_attributes


class BaseActionSetGroup(metaclass=ActionSetDeclarativeMetaclass):
    """
    Base raw class for creating custom Action Sets Groups.
    """

    def __init__(self):
        self.action_sets = ActionSetDict(deepcopy(self.base_action_sets))

    def __getitem__(self, name):
        return self.action_sets[name]

    def __iter__(self):
        for _, action_set in self.action_sets.items():
            yield action_set


class ActionSetGroup(BaseActionSetGroup):
    """
    Main class for grouping actions and use in views.
    """

    def __init__(self, request: HttpRequestType):
        self.request = request
        super().__init__()
        self.instantiate_action_sets()

    def instantiate_action_sets(self) -> None:
        for name, action_set in self.action_sets.items():
            self.action_sets.instantiate(name, request=self.request, action_set_group=self)


ActionSetGroupType = TypeVar('ActionSetGroupType', bound=ActionSetGroup)


class ActionSetWithPermissionCheck(ActionSet):
    """
    An example class of the action_set with permission protect.
    Uses HTML-element cleanup to prevent it's signature change in browser console.
    """

    elements_permissions: dict = {}  # {'object_name': [perm_1, perm_2,...]}
    protect_css_class: str = ''
    ignore_html_params: List[str] = ['type']

    def protect_element(self, element: ActionType) -> None:
        self.clean_css_classes(element)
        self.clean_html_params(element)

    def clean_html_params(self, element: ActionType) -> None:
        # clean html attributes.
        element.html_params = {
            name: value for name, value in element.html_params.items() if name in self.ignore_html_params
        }

    def clean_css_classes(self, element: ActionType) -> None:
        # clean js-* classes.
        element.css_classes = \
            [css_class for css_class in element.css_classes if not css_class.startswith('js')] + \
            [self.protect_css_class]

    def get_user_permissions(self) -> list:
        return []

    def get_element_permissions(self, element_name: str) -> list:
        return self.elements_permissions.get(element_name, [])

    def check_permissions(self) -> None:
        user_permissions = self.get_user_permissions()
        if user_permissions:
            for element_name, element in self.actions.items():
                element_permissions = self.get_element_permissions(element_name)
                if not all(permission in user_permissions for permission in element_permissions):
                    self.protect_element(element)

    def render(self) -> str:
        # verification of permissions before rendering.
        self.check_permissions()
        return super().render()


class ActionSetMenuMixin:
    """
    Миксин для реализации меню из Action Set.
    """

    url_attribute_name: str = 'data-url'
    activate_class: str = 'active'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.activate_menu_items()

    def activate_menu_items(self) -> None:
        current_url_name = resolve(self.request.path_info).url_name
        for name, action in self.actions.items():
            item_url_name = resolve(action.html_params[self.url_attribute_name]).url_name
            if item_url_name == current_url_name:
                action.css_classes.append(self.activate_class)


class BreadcrumbsSet(ActionSet):
    """
    Class for breadcrumbs realization.
    """

    crumb_class: Type[ActionType] = ActionButton
    active_crumb_css_class: str = ''

    def __init__(self, view_kwargs: dict, crumbs_set: List[str], *args, **kwargs):
        self.view_kwargs = view_kwargs
        self.crumbs_set = crumbs_set
        super().__init__(*args, **kwargs)
        self.reset_crumbs()

    def reset_crumbs(self):
        new_actions = OrderedDict()
        for crumb_name in self.crumbs_set:
            prepared_crumb_name = crumb_name.replace('*', '')
            new_actions[prepared_crumb_name] = self.actions[prepared_crumb_name]
            if '*' in crumb_name:
                new_actions[prepared_crumb_name].css_classes.append(self.active_crumb_css_class)
        self.actions = new_actions


BreadcrumbsSetType = TypeVar('BreadcrumbsSetType', bound=BreadcrumbsSet)



from typing import Type, List

from action_set.action_set import ActionSetType, ActionSetGroupType, BreadcrumbsSetType


class ActionSetMixinView:
    """
    A view mixin for adding action_set to views.
    """

    action_set_class: Type[ActionSetType] = None
    action_set_context_name: str = 'action_set'

    def get_action_set_kwargs(self, **kwargs) -> dict:
        return dict(request=self.request, **kwargs)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        if self.action_set_class:
            context.update({
                self.action_set_context_name: self.get_action_set_object()
            })
        return context

    def get_action_set_object(self) -> ActionSetType:
        return self.action_set_class(**self.get_action_set_kwargs())


class ActionSetGroupMixinView:
    """
    A view mixin for adding action_set_group to views.
    """

    action_set_group_class: Type[ActionSetGroupType] = None
    action_set_group_context_name: str = 'action_set_group'

    def get_action_set_group_kwargs(self, **kwargs) -> dict:
        return dict(request=self.request, **kwargs)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        if self.action_set_group_class:
            context.update({
                self.action_set_group_context_name: self.get_action_sets_group_object()
            })
        return context

    def get_action_sets_group_object(self) -> ActionSetGroupType:
        return self.action_set_group_class(**self.get_action_set_group_kwargs())


class BreadcrumbsMixinView:
    """
    Mixin for adding breadcrumbs to views.
    """

    breadcrumbs_set_class: Type[BreadcrumbsSetType] = None
    breadcrumbs_set_context_name: str = 'breadcrumbs'
    breadcrumbs_set: List[str] = []

    def get_breadcrumbs_set_kwargs(self, **kwargs) -> dict:
        return dict(request=self.request, crumbs_set=self.breadcrumbs_set, view_kwargs=self.kwargs, **kwargs)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        if self.breadcrumbs_set_class:
            context.update({
                self.breadcrumbs_set_context_name: self.get_breadcrumbs_set_object()
            })
        return context

    def get_breadcrumbs_set_object(self) -> ActionSetType:
        return self.breadcrumbs_set_class(**self.get_breadcrumbs_set_kwargs())
